#import "XeeDirectorySource.h"
#import "XeeImage.h"
#import "XeeKQueue.h"

@implementation XeeDirectorySource

-(id)initWithDirectory:(NSString *)directoryname
{
	if(self=[super init])
	{
		imgref=dirref=nil;

		if([self scanDirectory:directoryname])
		{
			[self pickImageAtIndex:0];
			return self;
		}
	}
	[self release];
	return nil;
}

-(id)initWithFilename:(NSString *)filename
{
	if(self=[super init])
	{
		imgref=dirref=nil;

		[self addFile:[XeeFSRef refForPath:filename] sort:NO]; // make sure this file is in the list
		[self pickImageAtIndex:0];
		if([self scanDirectory:[filename stringByDeletingLastPathComponent]]) return self;
	}
	[self release];
	return nil;
}

-(id)initWithImage:(XeeImage *)image
{
	if(self=[super init])
	{
		imgref=dirref=nil;

		NSString *filename=[image filename];
		[self addFile:[XeeFSRef refForPath:filename] sort:NO]; // make sure this file is in the list
		[self setCurrentImage:image index:0];
		if([self scanDirectory:[filename stringByDeletingLastPathComponent]]) return self;
	}
	[self release];
	return nil;
}

-(void)dealloc
{
	[[XeeKQueue defaultKQueue] removeObserver:dirref];
	[[XeeKQueue defaultKQueue] removeObserver:imgref];
	[dirref release];
	[imgref release];

	[super dealloc];
}



-(int)capabilities { return XeeNavigationCapable|XeeRenamingCapable|XeeCopyingCapable|XeeMovingCapable|XeeDeletionCapable; }



-(BOOL)scanDirectory:(NSString *)directoryname
{
	dirref=[XeeFSRef refForPath:directoryname];
	if(!dirref) return NO;

	[self addFiles:[self readDirectory:dirref] sort:YES clear:NO]; 

	[self setIcon:[[NSWorkspace sharedWorkspace] iconForFile:directoryname]];
	[icon setSize:NSMakeSize(16,16)];

	[[XeeKQueue defaultKQueue] addObserver:self selector:@selector(directoryChanged:)
	ref:dirref flags:NOTE_WRITE|NOTE_DELETE|NOTE_RENAME];

	needsrefresh=NO;

	return YES;
}

-(NSArray *)readDirectory:(XeeFSRef *)ref
{
	NSMutableArray *res=[NSMutableArray array];
	NSString *directoryname=[ref path];
	NSFileManager *fm=[NSFileManager defaultManager];
	NSArray *dircontents=[fm directoryContentsAtPath:directoryname];
	NSArray *filetypes=[XeeImage allFileTypes];
	NSEnumerator *enumerator=[dircontents objectEnumerator];
	NSString *file;
	while(file=[enumerator nextObject])
	{
		NSString *path=[directoryname stringByAppendingPathComponent:file];
		NSDictionary *attrs=[fm fileAttributesAtPath:path traverseLink:YES];
		NSString *type=NSFileTypeForHFSTypeCode([attrs fileHFSTypeCode]);
		NSString *ext=[[file pathExtension] lowercaseString];

		if([filetypes indexOfObject:ext]!=NSNotFound
		||[filetypes indexOfObject:type]!=NSNotFound)
		[res addObject:[XeeFSRef refForPath:path]];
//		[self addFile:[XeeFSRef refForPath:path]];
	}

	return res;
}

-(void)setCurrentImage:(XeeImage *)image index:(int)index
{
	[[XeeKQueue defaultKQueue] removeObserver:imgref];
	[imgref release];
	imgref=nil;

	[super setCurrentImage:image index:index];

	if(image)
	{
//		imgref=[[XeeFSRef refForPath:[image filename]] retain];
		imgref=[[files objectAtIndex:index] retain];
		[[XeeKQueue defaultKQueue] addObserver:self selector:@selector(fileChanged:)
		ref:imgref flags:NOTE_WRITE|NOTE_DELETE|NOTE_RENAME];
	}
}



-(NSString *)directory { return [dirref path]; }



-(void)fileChanged:(XeeKEvent *)event
{
	int flags=[event flags];
	XeeFSRef *ref=[event ref];

	if(ref!=imgref) return; // probably doesn't happen

	if(flags&NOTE_WRITE)
	{
		// refresh
		int index=currindex;
		[self setCurrentImage:nil index:-1];
		[self pickImageAtIndex:index next:nextindex];
	}
	if(flags&NOTE_RENAME)
	{
		if([ref isValid]&&[[ref parent] isEqual:dirref])
		{
			[self sortFiles];
			[currimage triggerPropertyChangeAction];
		}
		else [self removeFile:ref];
	}
	if(flags&NOTE_DELETE)
	{
		[self removeFile:ref];
	}
}

-(void)directoryChanged:(XeeKEvent *)event
{
	int flags=[event flags];
	XeeFSRef *ref=[event ref];

	if(flags&NOTE_WRITE)
	{
		[self setNeedsRefresh:YES];
	}
	if(flags&NOTE_RENAME)
	{
		if(![ref isValid]) [self removeAllFiles];
		else [currimage triggerPropertyChangeAction];
	}
	if(flags&NOTE_DELETE)
	{
		[self removeAllFiles];
	}
}



-(void)setNeedsRefresh:(BOOL)refresh
{
	if(needsrefresh==refresh) return;

	needsrefresh=refresh;

	if(needsrefresh) [self performSelector:@selector(refresh) withObject:nil afterDelay:0];
}

-(void)refresh
{
	if(!needsrefresh) return;

	[self addFiles:[self readDirectory:dirref] sort:YES clear:YES];

	needsrefresh=NO;
}

@end
