#import "XeeFileSource.h"
#import "XeeImage.h"



NSComparisonResult XeeFileSorter(id file1,id file2,XeeFileSource *source);



@implementation XeeFileSource

-(id)init
{
	if(self=[super init])
	{
		currindex=nextindex=previndex=-1;
		currimage=nextimage=previmage=nil;
		loadingimage=nil;

		files=[[NSMutableArray array] retain];

		listlock=[[NSRecursiveLock alloc] init];
		loadlock=[[NSRecursiveLock alloc] init];
		loader_running=NO;
	}
	return self;
}

-(void)dealloc
{
	[files release];
	[listlock release];
	[loadlock release];

	[currimage release];
	[nextimage release];
	[previmage release];

	[super dealloc];
}

-(void)stop
{
	exiting=YES;
	[loadingimage stopLoading];
}



-(int)numberOfImages
{
	return [files count];
}

-(int)indexOfCurrentImage
{
	if(currindex<0) return NSNotFound;
	return currindex;
}

-(NSString *)representedFilename
{
	if(currindex<0) return nil;
	return [[files objectAtIndex:currindex] path];
}

-(NSString *)descriptiveNameOfCurrentImage
{
	if(currindex<0) return nil;
	return [[[files objectAtIndex:currindex] description] lastPathComponent];
}

-(int)capabilities { return XeeNavigationCapable; }



-(void)pickImageAtIndex:(int)index
{
	int count=[self numberOfImages];

	int next;
	if(index==currindex&&nextindex>=0) next=nextindex;
	else if(index==count-1) next=count-2;
	else if(index==0) next=1;
	else if(index<currindex) next=index-1;
	else next=index+1;

	[self pickImageAtIndex:index next:next];
}

-(void)pickImageAtIndex:(int)index next:(int)next
{
	[listlock lock];

	XeeImage *newcurrimage=nil,*newnextimage=nil;

	if(index<0 || index>=[files count]) index=-1;
	else newcurrimage=[[self imageAtIndex:index] retain];

	if(next<0 || next>=[files count]) next=-1;
	else newnextimage=[[self imageAtIndex:next] retain];

	[loadlock lock];

	[self setPreviousImage:currimage index:currindex];
	[self setCurrentImage:newcurrimage index:index];
	[self setNextImage:newnextimage index:next];

	[newcurrimage release];
	[newnextimage release];

	[listlock unlock];

	if(loadingimage&&loadingimage!=currimage) [loadingimage stopLoading];

	[self launchLoader];

	[loadlock unlock];

	[self triggerImageChangeAction:currimage];
}



-(void)addFile:(id)file sort:(BOOL)sort
{
	[self lockList];

	int index=[files indexOfObject:file];
	if(index!=NSNotFound)
	{
		// already added
		[self unlockListWithUpdates:NO];
		return;
	}

	[files addObject:file];

	if(sort) [files sortUsingFunction:(int (*)(id, id, void *))XeeFileSorter context:self];

	[self unlockListWithUpdates:sort];

	[self triggerNumberOfImagesChangeAction];
	if(currindex<0&&delegate) [self pickImageAtIndex:0 next:-1];
}

-(void)addFiles:(NSArray *)newfiles sort:(BOOL)sort clear:(BOOL)clear
{
	[self lockList];

	if(clear)
	{
		[files removeAllObjects];
		[files addObjectsFromArray:newfiles];
	}
	else
	{
		NSEnumerator *enumerator=[newfiles objectEnumerator];
		id file;
		while(file=[enumerator nextObject]) [self addFile:file sort:NO];
	}

	if(sort) [files sortUsingFunction:(int (*)(id, id, void *))XeeFileSorter context:self];

	[self unlockListWithUpdates:sort||clear];

	[self triggerNumberOfImagesChangeAction];
	if(currindex<0&&delegate) [self pickImageAtIndex:0 next:-1];
}

-(void)removeFile:(id)file
{
	[self lockList];

	int index=[files indexOfObject:file];
	if(index==NSNotFound)
	{
		[self unlockListWithUpdates:NO];
		return;
	}

	[files removeObjectAtIndex:index];

	[self unlockListWithUpdates:YES];

	[self triggerNumberOfImagesChangeAction];
}

-(void)removeAllFiles
{
	[self lockList];

	[files removeAllObjects];

	[self unlockListWithUpdates:YES];

	[self triggerNumberOfImagesChangeAction];
	[self triggerImageChangeAction:nil];
}

-(void)sortFiles
{
	[self lockList];

	[files sortUsingFunction:(int (*)(id, id, void *))XeeFileSorter context:self];

	[self unlockListWithUpdates:YES];

	[self triggerNumberOfImagesChangeAction]; // just to update the position display
}


-(void)lockList
{
	[listlock lock];

	prevfile=currfile=nextfile=nil;
	if(previndex>=0) prevfile=[[files objectAtIndex:previndex] retain];
	if(currindex>=0) currfile=[[files objectAtIndex:currindex] retain];
	if(nextindex>=0) nextfile=[[files objectAtIndex:nextindex] retain];
}

-(void)unlockListWithUpdates:(BOOL)updated
{
	if(updated)
	{
		previndex=[files indexOfObject:prevfile];
		currindex=[files indexOfObject:currfile];
		nextindex=[files indexOfObject:nextfile];

		[loadlock lock];

		if(previndex==NSNotFound)
		{
			if(loadingimage==previmage) [loadingimage stopLoading];
			[self setPreviousImage:nil index:-1];
		}
		if(nextindex==NSNotFound)
		{
			if(loadingimage==nextimage) [loadingimage stopLoading];
			[self setNextImage:nil index:-1];
		}
		if(currindex==NSNotFound)
		{
			if(loadingimage==currimage) [loadingimage stopLoading];
			[self setCurrentImage:nil index:-1];

			int count=[files count];
			if(count)
			{
				int index;
				for(index=0;index<count;index++)
				if(XeeFileSorter(currfile,[files objectAtIndex:index],self)!=NSOrderedDescending)
				break;

				if(index<count) [self pickImageAtIndex:index];
				else [self pickImageAtIndex:count-1];
			}
			else [self triggerImageChangeAction:nil];
		}

		[prevfile release];
		[currfile release];
		[nextfile release];

		// should handle loading image
		[loadlock unlock];
	}

	[listlock unlock];
}




-(XeeFSRef *)refForPath:(NSString *)path { return nil; }



-(XeeImage *)imageAtIndex:(int)index
{
	if(index<0) return nil;
	else if(index==currindex) return currimage;
	else if(index==nextindex) return nextimage;
	else if(index==previndex) return previmage;
	else
	{
		id file=[files objectAtIndex:index];
		XeeFSRef *ref;

		if([file isKindOfClass:[XeeFSRef class]]) ref=file;
		else
		{
			ref=[self refForPath:[file description]];
			if(!ref) return nil;
			[files replaceObjectAtIndex:index withObject:ref];
		}

		return [XeeImage imageForRef:ref];
	}
}

-(void)setCurrentImage:(XeeImage *)image index:(int)index
{
	currindex=index;
	if(image==currimage) return;
	[currimage release];
	currimage=[image retain];
}

-(void)setPreviousImage:(XeeImage *)image index:(int)index
{
	previndex=index;
	if(image==previmage) return;
	[previmage release];
	previmage=[image retain];
}

-(void)setNextImage:(XeeImage *)image index:(int)index
{
	nextindex=index;
	if(image==nextimage) return;
	[nextimage release];
	nextimage=[image retain];
}



-(void)launchLoader
{
	if(!loader_running)
	{
		[NSThread detachNewThreadSelector:@selector(loader) toTarget:self withObject:nil];
		loader_running=YES;
	}
}

-(void)loader
{
	NSAutoreleasePool *pool=[[NSAutoreleasePool alloc] init];

	[self retain];

	[NSThread setThreadPriority:0.1];

	[loadlock lock];

	for(;;)
	{
		if(exiting) break;

		NSAutoreleasePool *pool=[[NSAutoreleasePool alloc] init];

		if(currimage&&[currimage needsLoading]) loadingimage=currimage;
		else if(nextimage&&[nextimage needsLoading]) loadingimage=nextimage;
		else break;

		[loadingimage retain];
		[loadlock unlock];

		double starttime=XeeGetTime();
		[loadingimage runLoader];
		double endtime=XeeGetTime();

		NSLog(@"%@: %g s",[loadingimage descriptiveFilename],endtime-starttime);

		[loadlock lock];
		[loadingimage release];

		[pool release];
	}

	loader_running=NO;
	loadingimage=nil;

	[loadlock unlock];

	[self release];

	[pool release];
}



@end



NSComparisonResult XeeFileSorter(id file1,id file2,XeeFileSource *source)
{
	return [[file1 description] compare:[file2 description] options:NSCaseInsensitiveSearch|NSNumericSearch];
}

